/*    File:  brelaz.c
 *  Author:  Peter Ross
 * Created:  28 May 1997
 * Updated:  4 June 1997
 *
 * Version 1.2
 * A bug-fix in backtracking (hugot: for(a=n->nArcs...)
 * Added stuff to tie-break by number of instances of arcs, not just
 *   number of arcs.
 *
 * Customisation notes are at the end of this introduction.
 *
 * This program runs the Brelaz algorithm to colour a graph. The graph
 * is provided on standard input or in a file, in PGA's timetable format.
 * The key ingredients are a line
 *     events N   # indicating how many events there are. They
 *                # are assumed to be numbered from 0 upward.
 * and lines of the form
 *     separate N1 N2 N3..
 * which say that events (ie nodes) N1, N2, N3.. must all be of different
 * colours. Thus a line
 *     separate 3 17
 * describes one edge, and a line
 *     separate 22 27 31 38 43
 * describes 5x4/2=10 edges, because it describes a clique of size 5.
  *
 * The program, by default, outputs something looking like a PGA
 * timetable chromosome. That is, it outputs a one-line string of characters.
 * The first character describes the colour of node 0, as '0'+N, where N
 * is the colour. So a character '9' indicates the 9th colour, and a ';'
 * indicates the 11th colour (since ';'-'0'=11 in ASCII terms).
 *   Such a chromosome can be fed into PGA's stand-alone decode program, eg
 *     brelaz foo.pga | decode -t -Ffoo.pga
 * to recover details of which events(nodes) are associated with which
 * timeslots(colours).
 *
 * Alternatively you can use the -c flag, as in
 *     brelaz -c foo.pga
 * to get output consisting of lines of the form
 *     N C  C1 C2 C3...
 * which mean that node N has colour C, but that node could also have had
 * colours C1 or C2 or C3 or.. without problems. This helps to show how
 * constrained each choice was. You might even use this information to
 * try adjusting the colouring, though of course you can only change a
 * single node before recomputing what the alternative choices might be.
 * For example, suppose you had 4 colours (0-3) to colour a graph consisting
 * of just one edge, between nodes 0 and 1. Then this output format would
 * give you:
 *     0 0  2 3
 *     1 1  2 3
 * You could then recolour node 0 as 2 or 3, but that would free up colour 0
 * for node 1 and disallow one of the other choices for node 1.
 *
 * The -p flag causes the program to also output a table of pairs of
 * colours (in lower triangular form) showing how many edges connect a
 * node of colour A to a node of colour B, for all A and B. This may be
 * useful if, for example, you need to consider permuting the colours
 * for some reason.
 *
 * The Brelaz algorithm chooses nodes to colour according to their
 * "colour degree", ie how many different colours are represented among
 * the already-coloured neighbours. If a node has 6 different colours
 * among its neighbours already, then it is more important to colour that
 * than another node which only has 5 among its neighbours. Ties are broken
 * according to node degree in the standard version.
 *
 * The -i flag causes colour-degree tie-breaks to be based on the
 * number of instances of arcs coming from a node, rather than the
 * number of distinct arcs (ie number of other nodes connected to it).
 *
 * Often, there will be several nodes of highest colour degree at any stage.
 * This program can backtrack through those choices as necessary. The
 * function reportColouring() below outputs the final colouring, perhaps
 * after backtracking. If you arrange for the function to call exit(), the
 * program ends. If you arrange for it to call return instead, it will
 * backtrack through the remaining choices.
 *
 * The implementation uses an array of node-description structures; element
 * K of that array describes the Kth node. But these structures also contain
 * forward and backward node pointers, so that they can be arranged into a
 * queue -- the queue of nodes still needing to be coloured at each stage.
 * This leads to reasonable space and time performance. For example, the
 * graph arising from the King Fahd University timetabling problem has 5,893
 * edges and can be coloured in about 4 seconds on my 166MHz (nonMMX) Pentium.
 * The Carleton University 91 problem has 29,312 edges and takes around 10
 * seconds.
 *
 * Customising:
 *
 * If you want to customise this program you should look at the function
 * greaterEqual(n1,n2). This defines the ordering of the queue (greatest
 * first) and is the ONLY function used to determine that ordering.
 *
 * The function colourNodes(L) does the job of working through the queue.
 * It is recursive, and handles the backtracking in a simple way.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "brelaz.h"

#define TRUE  1
#define FALSE 0
#define NONE -1         /* Flag value: `colour' of node not yet coloured */
                
int n_nodes;            /* Number of nodes. They are numbered from 0 */
int n_colours;          /* Number of colours available: numbered from 0 */
int n_days;             /* Used if trying to spread out colours */
int n_seats;            /* Used if limiting sizes */
struct node *nodeData;  /* The array of node structures */
struct node *queue;     /* The queue of nodes awaiting colouring */
int *seats_used;        /* Number of seats (arc-instances) used per colour */
int outputColours;      /* If FALSE, output pga chromo, else big table */
int outputPairs;        /* If TRUE, output colour-pairs table too */
int byInstances;        /* If TRUE, tie-break by #instances of arcs */
int debug;              /* If > 0, add extra output */

/***************************************************************/
/* Get the next number from the input: put it in the location  */
/* addressed by second argument. This function returns 0 on    */
/* EOF. If stopateol is true, it returns -1 when it hits \n    */
/* (after which some other procedure has to read past the \n), */
/* otherwise it continues looking for the next number.         */
/* A number has an optional sign, perhaps followed by digits,  */
/* perhaps followed by a decimal point, perhaps followed by    */
/* more digits. There must be a digit somewhere for it to count*/
/* as a number. So it would read any of:                       */
/*  -.5                                                        */
/*  -0.5                                                       */
/*  -.5.7                                                      */
/* as minus-a-half. In the last case, it would read .7 next    */
/* time around.                                                */
/*   There doesn't seem to be a neat and reliable way to do    */
/* all this, including stopateol, using scanf?                 */
/***************************************************************/

double getdouble(FILE * file, double *valaddr, int stopateol)
{
  int c;
  int found = FALSE, indecimal = FALSE;
  int sign = +1;
  double n = 0.0, p = 1.0;

  /* First find what looks like start of a number - the first digit. */
  /* And note any sign and whether we just passed a decimal point.   */
  do {
    c = fgetc(file);
    if (c == EOF)
      return (0);
    else if (stopateol && c == '\n')
      return (-1);
    else if (c == '+' || c == '-') {
      sign = (c == '+') ? +1 : -1;
      c = fgetc(file);
      if (c == EOF)
	return (0);
      else if (stopateol && c == '\n')
	return (-1);
    }
    if (c == '.') {
      indecimal = TRUE;
      c = fgetc(file);
      if (c == EOF)
	return (0);
      else if (stopateol && c == '\n')
	return (-1);
    }
    if (c >= '0' && c <= '9') {
      found = TRUE;
    } else {
      sign = +1;
      indecimal = FALSE;
    }
  } while (!found);

  /* Now we've got digit(s) ... */
  do {
    n = 10.0 * n + c - '0';
    p = 10.0 * p;
    c = fgetc(file);

    if ((c < '0') || (c > '9')) {
      found = FALSE;
      /* We've run out. If we already saw a decimal point, return now */
      if (indecimal) {
	if (c != EOF)
	  ungetc(c, file);
	*valaddr = sign * n / p;
	return (1);
      } else
	p = 1.0;
    }
  } while (found);

  /* We ran out and we didn't see a decimal point, so is this a decimal? */
  if (c != '.') {
    /* No, give it back to caller */
    if (c != EOF)
      ungetc(c, file);
    *valaddr = sign * n;
    return (1);
  } else {
    /* It is. Step past it, carry on hoping for more digits */
    c = fgetc(file);
    while (c >= '0' && c <= '9') {
      n = 10.0 * n + c - '0';
      p = p * 10.0;
      c = fgetc(file);
    }
    /* We've run out of digits but we have a number to give */
    if (c != EOF)
      ungetc(c, file);
    *valaddr = sign * n / p;
    return (1);
  }
}

/****************************************************/
/* getint(File, IntAddress, StopAtEol)              */
/* Use getdouble() above but convert result to int. */
/****************************************************/
int getint(FILE * f, int *valaddr, int stopateol)
{
  int r;
  double x;
  r = getdouble(f, &x, stopateol);
  *valaddr = (int) x;
  return (r);
}

/****************************************************/
/* check_initialised() exits if data structure not  */
/* yet initialised                                  */
/****************************************************/
void check_initialised(void)
{
  if(nodeData == (struct node *)NULL) {
    fprintf(stderr, "brelaz: didn't know enough to initialise structures");
    fprintf(stderr, " yet (eg how many nodes? how many colours?)\n");
    exit(1);
  }
}

/****************************************************/
/* setupNodeData() initialises the node array. Must */
/* know the number of nodes and colours to do so.   */
/****************************************************/
void setupNodeData(int nodes, int colours)
{
  int i, j;
  int *tmp;

  if(nodes == 0
     || colours == 0
     || nodeData != (struct node *)NULL) {
    return;
  }
  
 if((seats_used = (int *)malloc(n_colours*sizeof(int)))
    == (int *)NULL) {
   fprintf(stderr, "brelaz: cold not allocate seats_used\n");
   exit(1);
 }
 for(i=0; i<n_colours; i++) {
   seats_used[i] = 0;
 }
 
 if((nodeData = (struct node *)malloc(nodes*sizeof(struct node)))
     == (struct node *)NULL) {
    fprintf(stderr, "brelaz: not enough memory to alloc nodeData\n");
    exit(1);
  }

  nodeData[0].prev = (struct node *)NULL;
  
  nodeData[nodes-1].next = (struct node *) NULL;
  for(i=0;i<nodes;i++) {
    nodeData[i].id        = i;
    nodeData[i].colour    = NONE;
    nodeData[i].maxArcs   = CHUNK;
    nodeData[i].nArcs     = 0;
    nodeData[i].nOccurs   = 0;
    nodeData[i].nStudents = 0;
    nodeData[i].colDegree = 0;
        /* set up nc array: */
    if((tmp = (int *)malloc(colours*sizeof(int)))
       == (int *)NULL) {
      fprintf(stderr, "brelaz: failed to alloc nc array for node %d\n",i);
      exit(1);
    }
    for(j=0;j<colours;j++) {
      tmp[j] = 0;
    }
    nodeData[i].nc = tmp;
        /* set up arcs array: */
    if((tmp = (int *)malloc(CHUNK*sizeof(int)))
       == (int *)NULL) {
      fprintf(stderr, "brelaz: failed to alloc arcs array for node %d\n",i);
      exit(1);
    }
    nodeData[i].arcs = tmp;
        /* set up chain pointers. All nodes are in the queue at start. */
    if(i < nodes-1) {
      nodeData[i].next = &nodeData[i+1];
    } 
    if(i > 0) {
      nodeData[i].prev = &nodeData[i-1];
    } 
  }
}

/****************************************************/
/* addArc(n1,n2) installs an arc between nodes n1   */
/* and n2, IFF it is not there already. Else no-op. */
/* CALLER checks argument validity.                 */
/****************************************************/
void addArc(int n1, int n2)
{
  int i, j;
  int *tmp;
  struct node *np;

  nodeData[n1].nOccurs++;
  nodeData[n2].nOccurs++;
  
      /* Fix up node n1: */
  np = &nodeData[n1];
  tmp = np->arcs;
  for(i=0;i<np->nArcs;i++) {
    if(tmp[i] == n2) {
      return;
    }
  }
  if(np->nArcs == np->maxArcs) {
    i = np->maxArcs + CHUNK;
    if((tmp = (int *)realloc(tmp, i*sizeof(int))) == (int *)NULL) {
      fprintf(stderr, "brelaz: realloc failed for arcs of node %d\n", n1);
      exit(1);
    }
    np->maxArcs = i;
  }
  tmp[np->nArcs++] = n2;
  np->arcs = tmp;

      /* Now fix up n2 too: */
  np = &nodeData[n2];
  tmp = np->arcs;
  if(np->nArcs == np->maxArcs) {
    i = np->maxArcs + CHUNK;
    if((tmp = (int *)realloc(tmp, i*sizeof(int))) == (int *)NULL) {
      fprintf(stderr, "brelaz: realloc failed for arcs of node %d\n", n1);
      exit(1);
    }
    np->maxArcs = i;
  }
  tmp[np->nArcs++] = n1;
  np->arcs = tmp;
}

/****************************************************/
/* read_datafile(FilePtr) reads a PGA-type file     */
/* describing a graph-colouring problem             */
/* Stuff in the file that doesn't relate to graph-  */
/* colouring gets skipped                           */
/****************************************************/
int read_datafile(FILE *f)
{
  char buffer[256];
  double *data;
  struct constraint *ctmp;
  struct tt_params *paramp;
  int i, j, k, n, d0, d1, data_max, errflag;
  

  errflag = 0;
  data_max = CHUNK;
  data = (double *) malloc(sizeof(double) * data_max);

  n_nodes = 0; /* Must change this when reading data! 0 is a flag */

  while (fscanf(f, " %s", buffer) != EOF) {
    /* We have a non-whitespace string, do we know it? */
    for (paramp = paramlist;
	 (paramp->param_type != END_LIST)
	 && strncmp(paramp->param, buffer, strlen(buffer));
	 paramp++);

    if (paramp->param_type == END_LIST) {	/* If we don't.. */
      fscanf(f, "%*[^\n]");	/* skip to end of line */
    } else {			/* Else we know it .. */
      if (paramp->args_wanted != ALL) {	/* Gather the numbers */
	if (paramp->args_wanted > data_max) {
	  data_max = paramp->args_wanted;
	  if ((data = (double *) realloc(data, sizeof(double) * data_max))
	      == (double *) NULL) {
	    fprintf(stderr, "brelaz: could not realloc reading file\n");
	    exit(1);
	  }
	}
	for (n = 0; n < paramp->args_wanted; n++) {	/* read the numbers wanted */
	  if (!getdouble(f, &data[n], FALSE)) {
	    fprintf(stderr, "brelaz: unexpected EOF in timetable data\n");
	    return (0);
	  }
	}
      } else {			/* Gather ALL the numbers on this line */
	n = 0;
	while ((j = getdouble(f, &data[n], TRUE)) > 0) {
	  /* Now increment n .. error to do it in above line, since */
	  /* getdouble will be called once too often (at EOL)       */
	  n++;
	  /* Will we need more space? */
	  if (n == data_max) {
	    data_max += CHUNK;
	    if ((data = (double *) realloc(data, sizeof(double) * data_max))
		== (double *) NULL) {
	      fprintf(stderr, "brelaz: failed to realloc reading file\n");
	      exit(1);
	    }
	  }
	}
      }				/* End of number gathering */
      /* NB: next scan for a keyword using fscanf above will go past */
      /* any further newlines, even if we did getdouble(..TRUE)      */
      /* At this point, n contains the number of numbers read in for */
      /* the current item.                                           */

      /* d0 and d1 are used purely to make the code a little easier to read */
      d0 = (int) data[0];

      /* Store the data according to what it is: */
      switch (paramp->param_type) {
      case DATA_EVENTS:
	n_nodes = d0;
	/* Now we know how many nodes. Try setting up the data */
        /* structure array describing the nodes -- may fail if */
        /* we don't yet know the number of colours.            */
        setupNodeData(n_nodes, n_colours);
	break;
      case DATA_SLOTS:
	n_colours = d0;
	/* Now we know how many colours. Try setting up the data */
        /* structure array describing the nodes -- may fail if   */
        /* we don't yet know the number of nodes.                */
        setupNodeData(n_nodes, n_colours);
	break;
      case CONSTRAIN_CLASH:
	check_initialised();
	if (n < 1) {
	  fprintf(stderr,
	       "brelaz: expected at least one number in a `%s' constraint\n",
		  paramp->param);
	  errflag++;
          break;
	}
        if(n > 1) {
  	  for (i = 0; i < n - 1; i++) {
	    for (j = i + 1; j < n; j++) {
	      d0 = data[i];
	      d1 = data[j];
	      if (d0 == d1) {
	        fprintf(stderr,
                        "brelaz: cannot separate event %d from itself\n", d0);
                errflag++;
	      }
              if(d0 >= n_nodes || d1 >= n_nodes) {
                fprintf(stderr, "brelaz: nodes must be numbered 0..%d",
                        n_nodes - 1);
                fprintf(stderr, " -- error at arc (%d,%d)\n", d0, d1);
                errflag++;
              }
              addArc(d0,d1);
	    }			/* end of j loop */
	  }			/* end of i loop */
        }
        for(i=0; i< n; i++) {
          if(n_nodes > data[i]) {
            nodeData[(int)data[i]].nStudents++;
          }
        }
	break;
      case DATA_DAY:
        if((n_days = d0) < 1) {
          fprintf(stderr, "brelaz: crazy days (%d)\n", d0);
          exit(1);
        }
        break;
      case DATA_SEATS:
        if((n_seats = d0) < 1) {
          fprintf(stderr, "brelaz: crazy seats (%d)\n", d0);
          exit(1);
        }
        break;
      case DATA_SPREAD:
      case DATA_CLASHTYPE:
      case PENALTY_CLASH:
      case PENALTY_ORDER:
      case PENALTY_EXCLUDE:
      case PENALTY_NEARCLASH:
      case CONSTRAIN_BEFORE:
      case CONSTRAIN_EXCLUDE:
      case CONSTRAIN_PRESET:
	break;
      default:
	fprintf(stderr, "brelaz: uncatered-for type in the data\n");
	errflag++;
      }				/* end of param switch */
    }				/* end of dealing with one param */
  }				/* end of reading file */

  if(errflag) {
    exit(1);
  }
  
  free(data);

  return (1);
}

/****************************************************/
/* greaterEqual(NodePtr1, NodePtr2) returns TRUE if */
/* node1 precedes or equals node2 in standard order */
/****************************************************/
int greaterEqual(struct node *np1, struct node * np2)
{
  if((np1->colDegree) > (np2->colDegree)) {
    return(TRUE);
  }
  if((np1->colDegree) == (np2->colDegree)) {
    if(byInstances && ((np1->nOccurs) >= (np2->nOccurs))) {
      return(TRUE);
    }
    if((!byInstances) && ((np1->nArcs) >= (np2->nArcs))) {
      return(TRUE);
    }
  }
  return(FALSE);
}

/****************************************************/
/* swap(*n1,*n2) adjusts the pointers to            */
/* swap the places of n1 and n2 in the queue. Used  */
/* in sorting the queue.                            */
/****************************************************/
void swap(struct node *n1, struct node *n2)
{
  struct node *n1Next, *n2Next, *n1Prev, *n2Prev;
  struct node **preN1next, **postN2prev;
  
  if(queue == n1) {
    queue = n2;
  }

  n1Prev = n1->prev;
  n1Next = n1->next;
  n2Prev = n2->prev;
  n2Next = n2->next;
  if(n1Prev != (struct node *)NULL) {
    preN1next = &(n1Prev->next);
  }
  if(n2Next != (struct node *)NULL) {
    postN2prev = &(n2Next->prev);
  }

  n1->next = n2Next;
  n2->prev = n1Prev;
  if(n1Next == n2) {
    n2->next = n1;
  } else {
    n2->next = n1Next;
  }
  if(n2Prev == n1) {
    n1->prev = n2;
  } else {
    n1->prev = n2Prev;
  }
  if(n1Prev != (struct node *)NULL) {
    *preN1next = n2;
  }
  if(n2Next != (struct node *)NULL) {
    *postN2prev = n1;
  }
}

/****************************************************/
/* deQueue(nodePtr) removes that node from the      */
/* queue and return a pointer to what preceded      */
/* it in case we want to put it back later.         */
/****************************************************/
struct node *deQueue(struct node *n)
{
  struct node *prev;

  if((prev = n->prev) == (struct node *)NULL) {
    queue = n->next;
    if(n->next != (struct node *)NULL) {
      n->next->prev = (struct node *)NULL;
    }
  } else {
    prev->next = n->next;
    if(n->next != (struct node *)NULL) {
      n->next->prev = prev;
    }
  }
  return(prev);
}

/****************************************************/
/* enQueue(nodePtr, prevPtr) puts node back into    */
/* the queue right after prev                       */
/****************************************************/
void enQueue(struct node *n, struct node *prev)
{
  if(prev == (struct node *)NULL) {
    n->prev = (struct node *)NULL;
    n->next = queue;
    if(queue != (struct node *)NULL) {
      queue->prev = n;
    }
    queue = n;
  } else {
    n->next = prev->next;
    n->prev = prev;
    if(n->next != (struct node *)NULL) {
      n->next->prev = n;
    }
    prev->next = n;
  }
}
    
/****************************************************/
/* moveUp(i) moves node number i up the queue as    */
/* necessary, its colDegree has gone up.            */
/****************************************************/
void moveUp(int i)
{
  struct node *iptr, *n;
  
  iptr = &nodeData[i];
      /* Seek up the queue for something at least as good; stop there */
  for(n = iptr->prev; n != (struct node *)NULL; n = n->prev) {
    if(greaterEqual(n, iptr)) {
      break;
    }
  }
  if(n == iptr->prev) { /* Stopped next door? Do nothing */
    return;
  } else {
    (void) deQueue(iptr);
    enQueue(iptr, n);
  }
}

/****************************************************/
/* seatable(colour,seats) returns TRUE if colour c  */
/* can accomodate that many more                    */
/****************************************************/
int seatable(int colour, int seats)
{
  if(seats_used[colour]+seats < n_seats+1)
    return(TRUE);
  else
    return(FALSE);
}

/****************************************************/
/* colourIt(nodePtr) colours that node, and updates */
/* all its neighbours' colour arrays etc, and does  */
/* the necessary queue position adjustments.        */
/****************************************************/
int colourIt(struct node *n)
{
  int a, c, i, seats;

  seats = n->nStudents;
      /* Find a free colour for it */
  if(byInstances) {
    for(c=0; (c<n_colours) && !(n->nc[c]==0 && seatable(c,seats)); c = c+3);
    if(!(c < n_colours))
      for(c=2; (c<n_colours) && !(n->nc[c]==0 && seatable(c,seats)); c = c+3);
    if(!(c < n_colours))
      for(c=1; (c<n_colours) && !(n->nc[c]==0 && seatable(c,seats)); c = c+3);
  } else {
    for(c=0; c<n_colours && !(n->nc[c]==0 && seatable(c,seats)); c++);
  }

  if(!(c < n_colours)) {
    return(FALSE);
  }
  
  n->colour = c;
  seats_used[c] += seats;
  
      /* Go tell the neighbours all about it! */
  for(a=0; a < n->nArcs; a++) {
    i = n->arcs[a];
    if(nodeData[i].nc[c]++ == 0) {
      nodeData[i].colDegree++;
      if(nodeData[i].colour == NONE) {
        moveUp(i);
      }
    }
  }
  return(TRUE);
}

/****************************************************/
/* moveDown(i) moves node number i down the queue   */
/* as needed, its colDegree has gone down.          */
/****************************************************/
void moveDown(int i)
{
  struct node *iptr, *n, *m;

  m = iptr = &nodeData[i];
  for(n = iptr->next; n != (struct node *)NULL; n=n->next) {
    if(!greaterEqual(n, iptr)) {
      break;
    }
    m = m->next;
  }
  if(n == iptr->next) {
    return;
  } else {
    (void) deQueue(iptr);
    enQueue(iptr,m);
  }
}

/****************************************************/
/* uncolourIt(nodePtr) undoes colouring and updates */
/* all its neighbours' colour arrays etc, and does  */
/* the necessary queue position adjustments.        */
/****************************************************/
void uncolourIt(struct node *n)
{
  int a, i, c;

  c = n->colour;
  n->colour = NONE;
  seats_used[c] -= n->nStudents;
  
      /* Go tell the neighbours all about it (reverse order). */
  for(a=n->nArcs-1; a ; a--) {
    i = n->arcs[a];
    nodeData[i].nc[c]--;
    if(nodeData[i].nc[c] == 0) {
      nodeData[i].colDegree--;
      if(nodeData[i].colour == NONE) {
        moveDown(i);
      }
    }
  }
}

/****************************************************/
/* outputPairsTable() outputs a diagonal table in   */
/* which entry (i,j) shows the number of arcs that  */
/* join a node coloured i and another coloured j.   */
/****************************************************/
void outputPairsTable()
{
  int c, i, j, k, maxColour;
  int *tmp;

  maxColour = 0;
  if((tmp = (int *)malloc(n_colours*n_colours*sizeof(int)))
     == (int *)NULL) {
    fprintf(stderr,"brelaz: could not allocate pairs array\n");
    exit(1);
  }
  for(k=0;k<n_colours*n_colours;k++) {
    tmp[k] = 0;
  }
  for(i=0;i<n_nodes;i++) {
    if((c = nodeData[i].colour) > maxColour) {
      maxColour = c;
    }
    for(j=0;j<nodeData[i].nArcs;j++) {
      k=nodeData[nodeData[i].arcs[j]].colour;
      if(k<c) {
        tmp[c+n_colours*k]++;
      }
    }
  }
  printf("# pairs table: colours = 0 - %d\n   *\n", maxColour);
  for(c=1;c<=maxColour;c++) {
    for(k=0;k<c;k++) {
      printf(" %3d",tmp[c+n_colours*k]);
    }
    printf("   *\n");
  }
  free(tmp);
}

/****************************************************/
/* reportColouring() reports the final result. If   */
/* you do not make this exit, then the program will */
/* proceed to backtrack through other legitimate    */
/* Brelaz decision points.                          */
/****************************************************/
void reportColouring()
{
  int i, j;

  if(outputColours) {
        /* Print table: each line shows nodeId, colour, other choices */
    for(i=0; i<n_nodes; i++) {
      printf("%5d %3d  ", i, nodeData[i].colour);
      for(j=0; j<n_colours; j++) {
        if(nodeData[i].nc[j] == 0 && j != nodeData[i].colour) {
          printf(" %d", j);
        }
      }
      printf("\n");
    }
  } else {
        /* This outputs a PGA-type chromosome. You could feed */
        /* it into the   decode -t   program.                 */
    for(i=0; i<n_nodes; i++) {
      printf("%c", '0'+nodeData[i].colour);
    }
    printf("\n");
  }
  if(outputPairs) {
    outputPairsTable();
  }
  exit(0);
}


/****************************************************/
/* colourNodes() does the main job of colouring.    */
/* The argument indicates how many nodes are left   */
/* to colour.                                       */
/****************************************************/
void colourNodes(int left)
{
  int cd;
  struct node *n, *prev;
  if(!left) {
        /* If none left, output details. This may also call exit().. */
    reportColouring();
        /* ..but if it doesn't, return so backtracking can proceed.  */
    return;
  }
      /* Note colour degree of front of queue. We backtrack only     */
      /* through those of same colour degree. Otherwise it would     */
      /* not be Brelaz, and it would backtrack ultimately through    */
      /* all possible orderings of nodes (ouch).                     */
      /*   If colour degree is maximal we must go back a level..     */
  if((cd = queue->colDegree) == n_colours) {
    return;
  }
      /* ..else we can try colouring any of those that have top cd   */
  for(n=queue; n != (struct node *)NULL && cd == n->colDegree; n = n->next) {
        /* Take one off queue: queue length then (left-1) .. */
    prev = deQueue(n);
        /* .. do all that's necessary to colour it, if possible .. */
    if(colourIt(n)) {
          /* .. and recurse. */
      colourNodes(left - 1);
          /* If that returned we are backtracking, so undo the colouring .. */
      printf("-- backtrack at queue length %d\n", left);
      uncolourIt(n);
    }
        /* and reinstate n in the queue before proceeding with the loop. */
    enQueue(n,prev);
  }
}

/****************************************************/
/* printQueue() is a utility for printing out the   */
/* queue -- useful for debugging etc                */
/****************************************************/
void printQueue()
{
  struct node *n;
  for(n=queue; n != (struct node *)NULL; n=n->next) {
    printf("%4d: colDegree %-4d   degree %-4d   nOccurs %-4d   nStud %-4d\n",
           n->id, n->colDegree, n->nArcs, n->nOccurs, n->nStudents);
  }
}

main(int argc, char **argv)
{
  int i, j, swapped;
  FILE *f;
  struct node *n;

  argc--;
  argv++;

  debug = 0;
    
  outputColours = FALSE;
  outputPairs = FALSE;
  byInstances = FALSE;
  n_seats = (1<<30);  /* huge number in case user doesn't specify it */

  while(argv[0][0]=='-') {
    switch(argv[0][1]) {
        case 'd':
          debug++;
          break;
        case 'c':
          outputColours = TRUE;
          break;
        case 'p':
          outputPairs = TRUE;
          break;
        case 'i':
          byInstances = TRUE;
          break;
        default:
          fprintf(stderr,"usage: brelaz [-c] [-p] [-i] [-d] [file]\n");
          fprintf(stderr,"   -c = output verbose table, not PGA chromosome\n");
          fprintf(stderr,"   -p = output colour-pairs table too\n");
          fprintf(stderr,"   -i = tie-break by no. of instances of arcs\n");
          fprintf(stderr,"   -d = debug: extra output\n");
          fprintf(stderr,"   data from file or from stdin in PGA format\n");
          exit(1);
          break;
    }
    argc--;
    argv++;
  }
  
  if(argc == 0) {
    f = stdin;
  } else {
    if ((f = fopen(argv[0], "r")) == (FILE *) NULL) {
      fprintf(stderr, "brelaz: could not open data file %s\n",
	      argv[0]);
      exit(1);
    }
    
  }

  nodeData = (struct node *)NULL;
      /* Read the data file or stdin. After this, nodeData points */
      /* to the array of nodes.                                   */
  read_datafile(f);
      /* The queue is the chain of nodes to srat with. Needs to   */
      /* be sorted (by colour degree and then node degree).       */
  queue = nodeData;

  if(f != stdin) {
    fclose(f);
  }

      /* Sort the queue: very crudely by bubblesort: */
  swapped = TRUE;
  while(swapped) {
    swapped = FALSE;
    for(n=queue; !swapped && n->next != (struct node *)NULL; n = n->next) {
      if(!greaterEqual(n,n->next)) {
        swap(n,n->next);
      	swapped = TRUE;
      }
    }
  }
  if(debug)
    printQueue();
  
  /* This is where the real work starts.                               */
  /* The idea is to take a node off the front of the queue, colour it, */
  /* and adjust the queue-positions of all its neighbours. After that, */
  /* recurse. To get bactracking the essential trick is to loop at     */
  /* each level: INSIDE the loop, you dequeue one node and recurse. If */
  /* the recursion returns you failed to colour the graph, so requeue  */
  /* the node and undo the consequences of colouring it, and then loop.*/

  colourNodes(n_nodes);
  printf("no more cases to consider\n");
  
}

  
